"""
bigyo (비교: comparison)
========================

Python library for side-by-side diff in terminal.

use difflib as its engine.

Supports beautiful output with multi-width or zero-width character.

:date: 2022-12-15
:Author: LegenDUST
:Version: 0.1.1
"""

from bigyo._bigyo import *

__all__ = ['Bigyo']
