"""
bigyo.bigyo_renderer
====================

Renderer for Bigyo.

Inherit :class:`BigyoRenderer` to make your own renderer.
"""
from ._bigyo_renderer import *

__all__ = ['BigyoRenderer', 'SimpleBigyoRenderer', 'OnelineBigyoRenderer']
